#include "States.h"

States::States(WorldModel* worldModel)
{
	this->worldModel = worldModel;

	teamPosition = HALF_BACK;

	onGround = false;
}

void States::update()
{

	updateGoalFree();
	updateBallHolder();
	updateBallNearMe();
	updateOnGround();
	updateKickPossibility();
	updateFreeTeamMates();

	playMode = TPlayMode(worldModel->playMode);

}

void States::updateFreeTeamMates() {
	freeTeamMates.clear();
	for (int i = 0; i < worldModel->teammatesCount; i++) {
		if (isPathFreeToObject(&worldModel->teammates[i])) {
			freeTeamMates.push_back(worldModel->teammates[i]);
		}
	}
}

/**
	Method find out, if robot is able to kick to the ball, if ball is near him and if he is able
	to kick to the ball in some way.
	Method is NOT tested, beacause we were not been able to reach the ball in time of writing :).
*/
void States::updateKickPossibility() {
	CoordSpherical & ballCoords = worldModel->ball->position;
	double ballDistance = getDistanceFromMe(ballCoords);

	//if player can't reach the ball
	if (ballDistance < ballTouchDistance) {
		kickPossibility = NOT_POSSIBLE;
		return;
	}

	//left front kick
	if (isInHorizontalAngle(ballCoords, frontKickAngle, kickAngleTolerance)) {
		kickPossibility = LEFT_FRONT;
		return;
	}

	//right front kick
	if (isInHorizontalAngle(ballCoords, -frontKickAngle, kickAngleTolerance)) {
		kickPossibility = RIGHT_FRONT;
		return;
	}
	
	if (isInHorizontalAngle(ballCoords, 90, kickAngleTolerance)) {
		kickPossibility = LEFT_SIDE;
		return;
	}

	if (isInHorizontalAngle(ballCoords, -90, kickAngleTolerance)) {
		kickPossibility = RIGHT_SIDE;
		return;
	}

	kickPossibility = NOT_POSSIBLE;

	//TODO back kick
}

bool States::isInHorizontalAngle(const CoordSpherical & coords, double angle, double tolerance) {

	return isInAngle(coords.theta, angle, tolerance);
}

bool States::isInAngle(double testAngle, double expectedAngle, double tolerance) {
	if (testAngle > expectedAngle + tolerance ||
		testAngle < expectedAngle - tolerance) {
		return false;
	} else {
		return true;
	}
}

bool States::isInVerticalAngle(const CoordSpherical & coords, double angle, double tolerance) {
	return isInAngle(coords.phi, angle, tolerance);
}

/**
	Method checks, whether robot is staying or laying on ground.
	Method controls, if vertical angles are in normal range.
*/
void States::updateOnGround() {

	//CoordSpherical corner = getNearestCorner();

	double angle = 25;
	if (abs(worldModel->ourLeftCorner->position.phi) > angle) {
		onGround = true;
		/*cout << "*****\n*****\n*****\n*****\n*****" << endl;*/
	} else if(abs(worldModel->ourRightCorner->position.phi) > angle) {
		onGround = true;
		/*cout << "*****\n*****\n*****\n*****\n*****" << endl;*/
	} else if(abs(worldModel->opponentRightCorner->position.phi) > angle) {
		onGround = true;
		/*cout << "*****\n*****\n*****\n*****\n*****" << endl;*/
	} else if(abs(worldModel->opponentRightCorner->position.phi) > angle) {
		onGround = true;
		/*cout << "*****\n*****\n*****\n*****\n*****" << endl;*/
	} else {
		onGround = false;
		/*cout << "STANDING" << endl;*/
	}
}

/**
	returns vertical angle betwen object1, robot and object2
*/
double States::getAngleBetweenObjects(Object * object1, Object * object2) {
	return 180 + object1->position.phi + object2->position.phi;
}

CoordSpherical States::getNearestCorner() {
	CoordSpherical corners[4];
	corners[0] = worldModel->ourLeftCorner->position;
	corners[1] = worldModel->ourRightCorner->position;
	corners[2] = worldModel->opponentLeftCorner->position;
	corners[3] = worldModel->opponentRightCorner->position;
	
	double min = 100;
	int minIndex = 0;

	for (int i = 0; i < 4; i++) {
		if (corners[i].r < min) {
			min = corners[i].r;
			minIndex = i;
		}
	}

	cout << minIndex << "   ";

	return corners[minIndex];

}

void States::updateBallNearMe() {
	//if distance from ball is less than ballTouchDistance, I expect that ball is mine
	if (getDistanceFromMe(worldModel->ball->position) < ballTouchDistance)
	{
		ballNearMe = true;
	} else {
		ballNearMe = false;
	}
}

void States::updateGoalFree() {
	// first find center of the opponent goal
	double goalCenterTheta = (worldModel->opponentNet->left.position.theta +
		worldModel->opponentNet->right.position.theta) / 2;
	double goalCenterR = (worldModel->opponentNet->left.position.r +
		worldModel->opponentNet->right.position.r) / 2;
	CoordSpherical goalCenter =CoordSpherical(goalCenterR, goalCenterTheta, worldModel->opponentNet->left.position.phi);

	goalFree = isPathFreeToPosition(goalCenter);
	
}

bool States::isPathFreeToPosition(const CoordSpherical coords) {
	vector<CoordSpherical> blockingObjects = getBlockingObjectsCoords();

	bool freePath = true;

	vector<CoordSpherical>::iterator boIterator;
	
	for (boIterator = blockingObjects.begin(); boIterator != blockingObjects.end(); boIterator++) {
		if (isBetweenMeAndObject((CoordSpherical) *boIterator, coords)) {
			freePath = false;
			break;
		}
	}
	
	return freePath;
}

bool States::isPathFreeToObject(const Object * destObject) {

	return isPathFreeToPosition(destObject->position);
}

bool States::isBetweenMeAndObject(const CoordSpherical & blockingObjectCoords,
								  const CoordSpherical & destObjectCoords) {

	//if blocking object is behind destination object return false
	if (getDistanceFromMe(destObjectCoords) < getDistanceFromMe(blockingObjectCoords)) {
		return false;
	}
	if (blockingObjectCoords.theta > destObjectCoords.theta + angle ||
		blockingObjectCoords.theta < destObjectCoords.theta - angle) {
		return false;
	} else {
		return true;
	}
}

double States::getDistanceFromMe(const CoordSpherical & target)
{
	return target.r;
}

double States::getDistanceFromMe(const Object & target)
{
	return getDistanceFromMe(target.position);
}


/**
	Returns vectors of object, which are considered to be blocking path.
	Actually return opponents.
*/
vector<CoordSpherical> States::getBlockingObjectsCoords() {
	vector<CoordSpherical> blockingObjectsCoords;
	
	for (int i = 0; i < worldModel->opponentsCount; i++) {
		blockingObjectsCoords.push_back(worldModel->opponents[i].position);
	}
	
	return blockingObjectsCoords;
}
void States::updateBallHolder()
{
	//find distance from ball of the closest opponent
	double oppMinDistance = maxDistance;
	for (int i = 0; i < worldModel->opponentsCount; i++)
	{
		double distance = getDistance(worldModel->opponents[i].position, worldModel->ball->position);
		if (distance < oppMinDistance)
		{
			oppMinDistance = distance;
		}
	}

	//find distance from ball of the closest teamplayer
	double ourMinDistance = maxDistance;
	for (int i = 0; i < worldModel->teammatesCount; i++)
	{
		double distance = getDistance(worldModel->teammates[i].position, worldModel->ball->position);
		if (distance < ourMinDistance)
		{
			ourMinDistance = distance;
		}
	}

	//if some of the teammates or me is closer to the ball, we are considered to be ball holders
	if (ourMinDistance < oppMinDistance || getDistanceFromMe(worldModel->ball->position) < oppMinDistance)
	{
		ballHolder = OUR_TEAM;
	} else
	{
		ballHolder = OPPONENT_TEAM;
	}
}

double States::getDistance(const CoordSpherical & from, const CoordSpherical & to)
{
	CoordCartesian fromCart = CoordCartesian(from);
	CoordCartesian toCart = CoordCartesian(to);

	return getDistance(fromCart, toCart);
}

double States::getDistance(const CoordCartesian & from, const CoordCartesian & to)
{
	double xDiff = from.x - to.x;
	double yDiff = from.y - to.y;
	double zDiff = from.z - to.z;
	return sqrt(xDiff * xDiff + yDiff * yDiff + zDiff * zDiff);
}

double States::getDistance(const Object & from, const Object & to) {
	return getDistance(from.position, to.position);
}


bool States::isGoalFree() {
	return goalFree;
}

KickPossibility States::getKickPossibility() {
	return kickPossibility;
}

void States::setTeamPosition(TeamPosition _teamPosition)
{
	this->teamPosition = _teamPosition;
}

TeamPosition States::getTeamPosition() {
	return teamPosition;
}

vector<Player> States::getFreeTeamMates() {
	return freeTeamMates;
}

bool States::isOnGround() {
	return onGround;
}

States::~States()
{

}
